/**
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.bpmn2bpel;

import com.ibm.wsdl.extensions.soap.SOAPAddressImpl;
import com.ibm.wsdl.extensions.soap.SOAPBindingImpl;
import com.ibm.wsdl.extensions.soap.SOAPBodyImpl;
import com.ibm.wsdl.extensions.soap.SOAPOperationImpl;
import java.io.ByteArrayOutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.wsdl.Binding;
import javax.wsdl.BindingInput;
import javax.wsdl.BindingOperation;
import javax.wsdl.Definition;
import javax.wsdl.Import;
import javax.wsdl.Input;
import javax.wsdl.Message;
import javax.wsdl.Part;
import javax.wsdl.Port;
import javax.wsdl.PortType;
import javax.wsdl.Service;
import javax.wsdl.WSDLException;
import javax.wsdl.extensions.UnknownExtensibilityElement;
import javax.wsdl.extensions.soap.SOAPAddress;
import javax.wsdl.extensions.soap.SOAPBinding;
import javax.wsdl.extensions.soap.SOAPBody;
import javax.wsdl.extensions.soap.SOAPOperation;
import javax.wsdl.factory.WSDLFactory;
import javax.wsdl.xml.WSDLReader;
import javax.wsdl.xml.WSDLWriter;
import javax.xml.XMLConstants;
import javax.xml.namespace.QName;
import org.ow2.orchestra.bpmn2bpel.exceptions.TransformerException;
import org.ow2.orchestra.bpmn2bpel.generators.OnEventGenerator;
import org.ow2.orchestra.jaxb.bpmn.CorrelationPropertyRetrievalExpression;
import org.ow2.orchestra.jaxb.bpmn.Participant;
import org.ow2.orchestra.jaxb.bpmn.TCorrelationProperty;
import org.ow2.orchestra.jaxb.bpmn.TEndPoint;
import org.ow2.orchestra.jaxb.bpmn.TInterface;
import org.ow2.orchestra.jaxb.bpmn.TItemDefinition;
import org.ow2.orchestra.jaxb.bpmn.TMessage;
import org.ow2.orchestra.jaxb.bpmn.TProcess;
import org.ow2.orchestra.jaxb.bpmn.TUserTask;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.util.XmlUtil;
import org.w3c.dom.Document;
import org.w3c.dom.Element;


/**
 * @author Guillaume Porcher
 */
public final class WSDLGenerator {

  /**
   * Orchestra constant: replaced in orchestra engine with the server url.
   */
  private static final String ORCHESTRA_WS_URL = "${orchestra.ws.url}/";
  /**
   *
   */
  private static final String MSG_PART_NAME = "payload";

  private WSDLGenerator() {
  }

  /**
   * ws-addressing ns
   */
  public static final String XMLNS_WS_ADDRESSING = "http://www.w3.org/2005/08/addressing";
  /**
   * bpel 2.0 partnerLinkType ns.
   */
  public static final String XMLNS_BPEL_2_0_PLNKTYPE = "http://docs.oasis-open.org/wsbpel/2.0/plnktype";
  /**
   * bpel 2.0 varprop ns.
   */
  public static final String XMLNS_BPEL_2_0_VARPROP = "http://docs.oasis-open.org/wsbpel/2.0/varprop";
  /**
   * wsdl ns
   */
  public static final String XMLNS_WSDL = "http://schemas.xmlsoap.org/wsdl/";
  public static final String XMLNS_WSDL_SOAP = "http://schemas.xmlsoap.org/wsdl/soap/";

  private static final WSDLFactory WSDL_FACTORY;

  static {
    try {
      WSDL_FACTORY = WSDLFactory.newInstance();
    } catch (final WSDLException e) {
      throw new RuntimeException(e);
    }
  }

  public static Map<String, Definition> generateWsdlDefinitions(
          final BpmnMap bpmnMap, final QName processQName) throws WSDLException {
    final Definition wsdlDefinition = WSDLGenerator.WSDL_FACTORY.newDefinition();
    wsdlDefinition.setQName(processQName);
    wsdlDefinition.setTargetNamespace(processQName.getNamespaceURI());
    wsdlDefinition.addNamespace("tns", processQName.getNamespaceURI());
    wsdlDefinition.addNamespace("xsd", XMLConstants.W3C_XML_SCHEMA_NS_URI);
    wsdlDefinition.addNamespace("soap", WSDLGenerator.XMLNS_WSDL_SOAP);
    wsdlDefinition.setTypes(wsdlDefinition.createTypes());

    // get partner wsdl definitions
    final Map<String, Definition> partnerWSDLs = WSDLGenerator.getPartnerWSDLs(bpmnMap);

    // add imported wsdl definitions
    partnerWSDLs.putAll(WSDLGenerator.getImportedWSDLs(bpmnMap));

    if (!bpmnMap.getParticipants().isEmpty()) {
      // Add wsdl and bpel imports
      for (final Entry<String, Definition> partnerWSDL : partnerWSDLs.entrySet()) {
        final javax.wsdl.Import wsdlImport = wsdlDefinition.createImport();
        wsdlImport.setDefinition(partnerWSDL.getValue());
        wsdlImport.setNamespaceURI(partnerWSDL.getValue().getTargetNamespace());
        wsdlImport.setLocationURI(partnerWSDL.getKey());
        wsdlDefinition.addImport(wsdlImport);
      }

      // create partnerLinkType for participants
      WSDLGenerator.addPartnerLinkTypes(bpmnMap, wsdlDefinition);

    }

    WSDLGenerator.generateCorrelationProperties(bpmnMap, wsdlDefinition);
    partnerWSDLs.put("process.wsdl", wsdlDefinition);

    return partnerWSDLs;
  }


  public static void generateOnEventWsdlDefinition(
          final BpmnMap bpmnMap,
          final QName processQName,
          Set<String> onEventOperations,
          final Map<String, Definition> partnerWSDLs) {
    if (onEventOperations.isEmpty()) {
      return;
    }
    final Definition wsdlDefinition = WSDLGenerator.WSDL_FACTORY.newDefinition();
    wsdlDefinition.setQName(new QName(processQName.getNamespaceURI(), processQName.getLocalPart() + "InternalEvents"));
    wsdlDefinition.setTargetNamespace(processQName.getNamespaceURI());
    wsdlDefinition.addNamespace("tns", processQName.getNamespaceURI());
    wsdlDefinition.addNamespace("xsd", XMLConstants.W3C_XML_SCHEMA_NS_URI);
    wsdlDefinition.addNamespace("soap", WSDLGenerator.XMLNS_WSDL_SOAP);
    wsdlDefinition.setTypes(wsdlDefinition.createTypes());

    // generate PortType
    final PortType portType = wsdlDefinition.createPortType();
    portType.setQName(new QName(processQName.getNamespaceURI(), "internalOnEventPT", "tns"));
    portType.setUndefined(false);
    wsdlDefinition.addPortType(portType);

    // add soap binding for the PortType
    final Binding binding = wsdlDefinition.createBinding();
    binding.setQName(new QName(processQName.getNamespaceURI(), "internalOnEventPTSOAPBinding", "tns"));
    binding.setUndefined(false);
    binding.setPortType(portType);
    final SOAPBinding soapBinding = new SOAPBindingImpl();
    soapBinding.setStyle("document");
    soapBinding.setTransportURI("http://schemas.xmlsoap.org/soap/http");
    binding.addExtensibilityElement(soapBinding);
    wsdlDefinition.addBinding(binding);

    for (final String op : onEventOperations) {
      // add operations to portType
      final javax.wsdl.Operation operation = wsdlDefinition.createOperation();
      operation.setName(op);
      operation.setUndefined(false);
      portType.addOperation(operation);

      // add binding for operation
      final BindingOperation bindingOperation = wsdlDefinition.createBindingOperation();
      bindingOperation.setName(op);
      bindingOperation.setOperation(operation);
      binding.addBindingOperation(bindingOperation);
      final SOAPOperation soapOperation = new SOAPOperationImpl();
      soapOperation.setSoapActionURI(portType.getQName().getNamespaceURI() + portType.getQName().getLocalPart() + "/" + op);
      bindingOperation.addExtensibilityElement(soapOperation);

      final QName inMessageQName = new QName(processQName.getNamespaceURI(), OnEventGenerator.ON_EVENT_MSG_TYPE, "tns");
      // generate inMessage
      final Message message = wsdlDefinition.createMessage();
      message.setUndefined(false);
      message.setQName(inMessageQName);

      final Part messagePart = wsdlDefinition.createPart();
      // Process will use Document/literal binding
      // According to WS-Basic profile 1.1, messages contains only one part.
      messagePart.setName(WSDLGenerator.MSG_PART_NAME);

      messagePart.setTypeName(new QName(XMLConstants.W3C_XML_SCHEMA_NS_URI, "string", "xsd"));
      message.addPart(messagePart);
      wsdlDefinition.addMessage(message);
      final Input input = wsdlDefinition.createInput();
      input.setMessage(message);
      operation.setInput(input);

      final BindingInput bindingInput = wsdlDefinition.createBindingInput();
      final SOAPBody inputSoapBody = new SOAPBodyImpl();
      inputSoapBody.setUse("literal");
      bindingInput.addExtensibilityElement(inputSoapBody);
      bindingOperation.setBindingInput(bindingInput);
    }
    // add service and ports
    final Service service = wsdlDefinition.createService();
    final TProcess tProcess = bpmnMap.getProcesses().get(processQName);
    service.setQName(new QName(
            processQName.getNamespaceURI(),
            tProcess.getName() + "InternalService",
            processQName.getPrefix()));
    wsdlDefinition.addService(service);

    final Port port = wsdlDefinition.createPort();
    port.setBinding(binding);
    port.setName(binding.getPortType().getQName().getLocalPart() + "Port");

    service.addPort(port);

    final SOAPAddress soapAddress = new SOAPAddressImpl();
    port.addExtensibilityElement(soapAddress);
    soapAddress.setLocationURI(WSDLGenerator.ORCHESTRA_WS_URL + tProcess.getName() + "/" + port.getName());

    // add partnerlinktype to wsdl
    final UnknownExtensibilityElement partnerLinkTypeExtElement = new UnknownExtensibilityElement();
    final Document partnerLinkTypeDocument = XmlUtil.getNewDocument();
    final Element partnerLinkElt = partnerLinkTypeDocument.createElementNS(WSDLGenerator.XMLNS_BPEL_2_0_PLNKTYPE, "plnk:partnerLinkType");
    partnerLinkTypeDocument.appendChild(partnerLinkElt);
    partnerLinkElt.setAttribute("name", OnEventGenerator.ON_EVENT_PARTNER_LINK_TYPE);
    partnerLinkTypeExtElement.setElement(partnerLinkElt);
    wsdlDefinition.addExtensibilityElement(partnerLinkTypeExtElement);

    final Element roleElt = partnerLinkTypeDocument.createElementNS(WSDLGenerator.XMLNS_BPEL_2_0_PLNKTYPE, "plnk:role");
    partnerLinkElt.appendChild(roleElt);
    roleElt.setAttribute("name", OnEventGenerator.ON_EVENT_PARTNER_LINK_ROLE);
    roleElt.setAttribute("portType", "tns:" + portType.getQName().getLocalPart());

    // Add property to wsdl
    final UnknownExtensibilityElement propertyExtElement = new UnknownExtensibilityElement();
    final Document propertyDocument = XmlUtil.getNewDocument();
    final Element propertyElt = propertyDocument.createElementNS(WSDLGenerator.XMLNS_BPEL_2_0_VARPROP, "vprop:property");
    propertyDocument.appendChild(propertyElt);
    propertyElt.setAttribute("name", OnEventGenerator.ON_EVENT_PROPERTY);
    propertyElt.setAttribute("type", "xsd:string");
    propertyExtElement.setElement(propertyElt);
    wsdlDefinition.addExtensibilityElement(propertyExtElement);

    // Add propertyAlias to wsdl
    final UnknownExtensibilityElement propertyAliasExtElement = new UnknownExtensibilityElement();
    final Document propertyAliasDocument = XmlUtil.getNewDocument();
    final Element propertyAliasElt = propertyAliasDocument.createElementNS(WSDLGenerator.XMLNS_BPEL_2_0_VARPROP, "vprop:propertyAlias");
    propertyAliasDocument.appendChild(propertyAliasElt);
    propertyAliasElt.setAttribute("propertyName", "tns:" + OnEventGenerator.ON_EVENT_PROPERTY);
    propertyAliasElt.setAttribute("messageType", "tns:" + OnEventGenerator.ON_EVENT_MSG_TYPE);
    propertyAliasElt.setAttribute("part", "payload");
    propertyAliasExtElement.setElement(propertyAliasElt);
    wsdlDefinition.addExtensibilityElement(propertyAliasExtElement);

    // Add definition to list
    partnerWSDLs.put("processInternalEvents.wsdl", wsdlDefinition);

  }

  public static Definition generateTaskWsdlDefinition(
          final BpmnMap bpmnMap,
          final QName processQName,
          final TUserTask userTask,
          final Map<String, Definition> partnerWSDLs) {
    final Definition wsdlDefinition = WSDLGenerator.WSDL_FACTORY.newDefinition();
    wsdlDefinition.setQName(new QName(processQName.getNamespaceURI(), processQName.getLocalPart() + "Task" + userTask.getId()));
    wsdlDefinition.setTargetNamespace(processQName.getNamespaceURI());
    wsdlDefinition.addNamespace("tns", processQName.getNamespaceURI());
    wsdlDefinition.addNamespace("xsd", XMLConstants.W3C_XML_SCHEMA_NS_URI);
    wsdlDefinition.addNamespace("soap", WSDLGenerator.XMLNS_WSDL_SOAP);
    wsdlDefinition.setTypes(wsdlDefinition.createTypes());

    // generate PortType
    final PortType portType = wsdlDefinition.createPortType();
    portType.setQName(new QName(processQName.getNamespaceURI(), "takskPT" + userTask.getId(), "tns"));
    portType.setUndefined(false);
    wsdlDefinition.addPortType(portType);

    // add soap binding for the PortType
    final Binding binding = wsdlDefinition.createBinding();
    binding.setQName(new QName(processQName.getNamespaceURI(), "takskPT" + userTask.getId() + "SOAPBinding", "tns"));
    binding.setUndefined(false);
    binding.setPortType(portType);
    final SOAPBinding soapBinding = new SOAPBindingImpl();
    soapBinding.setStyle("document");
    soapBinding.setTransportURI("http://schemas.xmlsoap.org/soap/http");
    binding.addExtensibilityElement(soapBinding);
    wsdlDefinition.addBinding(binding);

    // add input operation to portType
    final javax.wsdl.Operation operation = wsdlDefinition.createOperation();
    operation.setName("inOperation");
    operation.setUndefined(false);
    portType.addOperation(operation);

    // add binding for operation
    final BindingOperation bindingOperation = wsdlDefinition.createBindingOperation();
    bindingOperation.setName("inOperation");
    bindingOperation.setOperation(operation);
    binding.addBindingOperation(bindingOperation);
    final SOAPOperation soapOperation = new SOAPOperationImpl();
    soapOperation.setSoapActionURI(portType.getQName().getNamespaceURI() + portType.getQName().getLocalPart() + "/inOperation");
    bindingOperation.addExtensibilityElement(soapOperation);

    final QName inMessageQName = new QName(processQName.getNamespaceURI(), "task" + userTask.getId() + "InMsg", "tns");
    // generate inMessage
    final Message message = wsdlDefinition.createMessage();
    message.setUndefined(false);
    message.setQName(inMessageQName);

    final Part messagePart = wsdlDefinition.createPart();
    // Process will use Document/literal binding
    // According to WS-Basic profile 1.1, messages contains only one part.
    messagePart.setName(WSDLGenerator.MSG_PART_NAME);

    messagePart.setTypeName(new QName(XMLConstants.W3C_XML_SCHEMA_NS_URI, "string", "xsd"));
    message.addPart(messagePart);
    wsdlDefinition.addMessage(message);
    final Input input = wsdlDefinition.createInput();
    input.setMessage(message);
    operation.setInput(input);

    final BindingInput bindingInput = wsdlDefinition.createBindingInput();
    final SOAPBody inputSoapBody = new SOAPBodyImpl();
    inputSoapBody.setUse("literal");
    bindingInput.addExtensibilityElement(inputSoapBody);
    bindingOperation.setBindingInput(bindingInput);

    // add output operation to portType
    final javax.wsdl.Operation outOperation = wsdlDefinition.createOperation();
    outOperation.setName("outOperation");
    outOperation.setUndefined(false);
    portType.addOperation(outOperation);

    // add binding for operation
    final BindingOperation outBindingOperation = wsdlDefinition.createBindingOperation();
    outBindingOperation.setName("outOperation");
    outBindingOperation.setOperation(outOperation);
    binding.addBindingOperation(outBindingOperation);
    final SOAPOperation outSoapOperation = new SOAPOperationImpl();
    outSoapOperation.setSoapActionURI(portType.getQName().getNamespaceURI() + portType.getQName().getLocalPart() + "/outOperation");
    outBindingOperation.addExtensibilityElement(outSoapOperation);

    final QName outMessageQName = new QName(processQName.getNamespaceURI(), "task" + userTask.getId() + "OutMsg", "tns");
    // generate inMessage
    final Message outMessage = wsdlDefinition.createMessage();
    outMessage.setUndefined(false);
    outMessage.setQName(outMessageQName);

    final Part outMessagePart = wsdlDefinition.createPart();
    // Process will use Document/literal binding
    // According to WS-Basic profile 1.1, messages contains only one part.
    outMessagePart.setName(WSDLGenerator.MSG_PART_NAME);

    outMessagePart.setTypeName(new QName(XMLConstants.W3C_XML_SCHEMA_NS_URI, "string", "xsd"));
    outMessage.addPart(outMessagePart);
    wsdlDefinition.addMessage(outMessage);
    final Input outOpInput = wsdlDefinition.createInput();
    outOpInput.setMessage(outMessage);
    outOperation.setInput(outOpInput);

    final BindingInput outBindingInput = wsdlDefinition.createBindingInput();
    final SOAPBody outOpInputSoapBody = new SOAPBodyImpl();
    outOpInputSoapBody.setUse("literal");
    outBindingInput.addExtensibilityElement(outOpInputSoapBody);
    outBindingOperation.setBindingInput(outBindingInput);

    // add service and ports
    final Service service = wsdlDefinition.createService();
    final TProcess tProcess = bpmnMap.getProcesses().get(processQName);
    service.setQName(new QName(
            processQName.getNamespaceURI(),
            tProcess.getName() + "InternalService",
            processQName.getPrefix()));
    wsdlDefinition.addService(service);

    final Port port = wsdlDefinition.createPort();
    port.setBinding(binding);
    port.setName(binding.getPortType().getQName().getLocalPart() + "Port");

    service.addPort(port);

    final SOAPAddress soapAddress = new SOAPAddressImpl();
    port.addExtensibilityElement(soapAddress);
    soapAddress.setLocationURI(WSDLGenerator.ORCHESTRA_WS_URL + tProcess.getName() + "/tasks/" + port.getName());

    // Add definition to list
    partnerWSDLs.put("task" + userTask.getId() + ".wsdl", wsdlDefinition);

    return wsdlDefinition;
  }

  /**
   * Get a map of WSDL definitions for definitions embedded in a endpoint.
   *
   * @param bpmnMap
   * @return
   * @throws WSDLException
   */
  private static Map<String, Definition> getPartnerWSDLs(final BpmnMap bpmnMap) throws WSDLException {
    final Map<String, Definition> partnerWsdls = new TreeMap<String, Definition>();
    for (final Participant participant : bpmnMap.getParticipants().values()) {
      for (final QName endPointQName : participant.getEndPointReves()) {
        final TEndPoint endPoint = bpmnMap.getEndpoints().get(endPointQName);
        final List<Element> endPointElements = endPoint.getExtensionElements().getAnies();
        for (final Element element : endPointElements) {
          if (element.getNamespaceURI().equals(WSDLGenerator.XMLNS_WS_ADDRESSING)
              && element.getLocalName().equals("EndpointReference")) {
            final Element metadataElt = XmlUtil.element(element, WSDLGenerator.XMLNS_WS_ADDRESSING, "Metadata");
            if (metadataElt != null) {
              final Element wsdlElt = XmlUtil.element(metadataElt, WSDLGenerator.XMLNS_WSDL, "definitions");
              if (wsdlElt != null) {
                final WSDLFactory wsdlFactory = WSDLFactory.newInstance();
                final WSDLReader wsdlReader = wsdlFactory.newWSDLReader();
                final Definition definition = wsdlReader.readWSDL((String) null, wsdlElt);
                partnerWsdls.put("partner" + partnerWsdls.size() + ".wsdl", definition);
              }
            }
          }
        }
      }
    }
    return partnerWsdls;
  }

  /**
   * Get a map of WSDL definitions for definitions imported by the BPMN file.
   *
   * @param bpmnMap
   * @return
   * @throws WSDLException
   */
  private static Map<String, Definition> getImportedWSDLs(final BpmnMap bpmnMap) throws WSDLException {
    return bpmnMap.getImportedWsdlDefinitions();
  }

  private static void addPartnerLinkTypes(final BpmnMap bpmnMap, final Definition wsdlDefinition) {
    for (final Entry<QName, Participant> entry : bpmnMap.getParticipants().entrySet()) {
      final QName participantQName = entry.getKey();
      final Participant participant = entry.getValue();

      // add partnerlinktype to wsdl
      final UnknownExtensibilityElement extElement = new UnknownExtensibilityElement();
      final Document document = XmlUtil.getNewDocument();
      final Element partnerLinkElt = document.createElementNS(WSDLGenerator.XMLNS_BPEL_2_0_PLNKTYPE, "plnk:partnerLinkType");
      document.appendChild(partnerLinkElt);
      partnerLinkElt.setAttribute("name", bpmnMap.getPartnerLinkTypeName(participantQName));
      extElement.setElement(partnerLinkElt);
      wsdlDefinition.addExtensibilityElement(extElement);

      for (final QName interfaceQName : participant.getInterfaceReves()) {
        final TInterface tInterface = bpmnMap.getInterfaces().get(interfaceQName);
        final Element roleElt = document.createElementNS(WSDLGenerator.XMLNS_BPEL_2_0_PLNKTYPE, "plnk:role");
        partnerLinkElt.appendChild(roleElt);
        roleElt.setAttribute("name", bpmnMap.getRoleName(interfaceQName));
        if (tInterface.getImplementationRef() != null) {
          roleElt.setAttribute(
                  "portType",
                  bpmnMap.getNamespaceAlias(tInterface.getImplementationRef().getNamespaceURI())
                  + ":" + tInterface.getImplementationRef().getLocalPart());
          roleElt.setAttributeNS(XMLConstants.XMLNS_ATTRIBUTE_NS_URI,
                  "xmlns:" + bpmnMap.getNamespaceAlias(tInterface.getImplementationRef().getNamespaceURI()),
                  tInterface.getImplementationRef().getNamespaceURI());
        } else {
          roleElt.setAttribute(
                  "portType",
                  bpmnMap.getNamespaceAlias(interfaceQName.getNamespaceURI()) + ":" + tInterface.getName());
          roleElt.setAttributeNS(XMLConstants.XMLNS_ATTRIBUTE_NS_URI,
                  "xmlns:" + bpmnMap.getNamespaceAlias(interfaceQName.getNamespaceURI()),
                  interfaceQName.getNamespaceURI());
        }
      }
    }
  }

  /**
   * Recursively write the list of WSDL definitions.
   *
   * @param partnerWSDLs the list of WSDL definitions to write, with the wanted file name.
   *                     Imported definitions will be generated too, with generated names.
   * @return a map of the wsdls generated with file name as key and byte content as value.
   * @throws WSDLException
   */
  public static Map<String, byte[]> writeWsdls(final Map<String, Definition> partnerWSDLs) throws WSDLException {
    final Map<String, byte[]> files = new TreeMap<String, byte[]>();

    final WSDLWriter wsdlWriter = WSDLGenerator.WSDL_FACTORY.newWSDLWriter();

    final Map<Definition, String> writtenWsdls = new HashMap<Definition, String>();
    for (final Definition partnerWSDL : partnerWSDLs.values()) {
      files.putAll(WSDLGenerator.writeWsdl(partnerWSDL, partnerWSDLs, writtenWsdls, wsdlWriter));
    }
    return files;
  }

  /**
   * Recursively write a WSDL definition.
   *
   * @param wsdlToWrite  the main wsdl to write
   * @param partnerWSDLs a map of wsdl with fixed names (wsdl directly used by the process).
   *                     For definitions in this map, the key will be used as wsdl file name.
   *                     For definitions not in this map, a new name will be generated.
   * @param writtenWsdls a map of the wsdl definitions already written, with the file name associated.
   * @param wsdlWriter   the writer used to serialize definitions to XML.
   * @return a map of the wsdls generated with file name as key and byte content as value.
   * @throws WSDLException
   */
  private static Map<String, byte[]> writeWsdl(
          final Definition wsdlToWrite,
          final Map<String, Definition> partnerWSDLs,
          final Map<Definition, String> writtenWsdls,
          final WSDLWriter wsdlWriter) throws WSDLException {
    // do nothing if definition already generated
    if (writtenWsdls.containsKey(wsdlToWrite)) {
      return Collections.emptyMap();
    }

    final Map<String, byte[]> result = new TreeMap<String, byte[]>();
    // get name for this definition
    final String wsdlName;
    if (partnerWSDLs.containsValue(wsdlToWrite)) {
      String name = null;
      for (final Map.Entry<String, Definition> entry : partnerWSDLs.entrySet()) {
        if (entry.getValue().equals(wsdlToWrite)) {
          name = entry.getKey();
          break;
        }
      }
      wsdlName = name;
    } else {
      wsdlName = Misc.getHumanReadableId("wsdl") + ".wsdl";
    }
    // add to the list of wsdl definitions processed
    writtenWsdls.put(wsdlToWrite, wsdlName);

    // recursively write imported definitions
    // and update location
    for (final Collection<Import> wsdlImportVector : (Collection<Collection<Import>>) wsdlToWrite.getImports().values()) {
      for (final Import wsdlImport : wsdlImportVector) {
        final Definition importedDefinition = wsdlImport.getDefinition();
        if (!writtenWsdls.containsKey(importedDefinition)) {
          result.putAll(WSDLGenerator.writeWsdl(importedDefinition, partnerWSDLs, writtenWsdls, wsdlWriter));
        }
        wsdlImport.setLocationURI(writtenWsdls.get(importedDefinition));
      }
    }

    // generate xml for this definition
    final ByteArrayOutputStream partnerWsdlOutput = new ByteArrayOutputStream();

    wsdlWriter.writeWSDL(wsdlToWrite, partnerWsdlOutput);

    // Indent WSDL output
    final String indentedOut = XmlUtil.toString(XmlUtil.getDocumentFromString(partnerWsdlOutput.toString(), false));

    // Add content to list of files
    result.put(wsdlName, indentedOut.getBytes());

    return result;
  }

  private static void generateCorrelationProperties(final BpmnMap bpmnMap, final Definition wsdlDefinition) {
    for (final TCorrelationProperty correlationProperty : bpmnMap.getCorrelationProperties().values()) {

      // add property to wsdl
      final UnknownExtensibilityElement extElement = new UnknownExtensibilityElement();
      final Document document = XmlUtil.getNewDocument();
      final Element propertyElt = document.createElementNS(WSDLGenerator.XMLNS_BPEL_2_0_VARPROP, "vprop:property");
      document.appendChild(propertyElt);
      propertyElt.setAttribute("name", correlationProperty.getId());
      TItemDefinition itemDefinition = bpmnMap.getItemDefinitions().get(correlationProperty.getType());
      propertyElt.setAttributeNS(XMLConstants.XMLNS_ATTRIBUTE_NS_URI,
              "xmlns:" + bpmnMap.getNamespaceAlias(itemDefinition.getStructureRef().getNamespaceURI()),
              itemDefinition.getStructureRef().getNamespaceURI());
      // TODO: check type to see if it's a type or an element
      propertyElt.setAttribute("type", bpmnMap.getNamespaceAlias(itemDefinition.getStructureRef().getNamespaceURI()) + ":" + itemDefinition.getStructureRef().getLocalPart());
      extElement.setElement(propertyElt);
      wsdlDefinition.addExtensibilityElement(extElement);

      for (CorrelationPropertyRetrievalExpression retrievalExpression : correlationProperty.getCorrelationPropertyRetrievalExpressions()) {
        // add property alias to wsdl
        final UnknownExtensibilityElement propAliasExtElement = new UnknownExtensibilityElement();
        final Document propAliasDocument = XmlUtil.getNewDocument();
        final Element propertyAliasElt = propAliasDocument.createElementNS(WSDLGenerator.XMLNS_BPEL_2_0_VARPROP, "vprop:propertyAlias");
        propAliasDocument.appendChild(propertyAliasElt);
        propertyAliasElt.setAttribute("propertyName", "tns:" + correlationProperty.getId());

        TMessage message = bpmnMap.getMessages().get(retrievalExpression.getMessageRef());
        TItemDefinition messageItemDefinition = bpmnMap.getItemDefinitions().get(message.getItemRef());

        propertyAliasElt.setAttributeNS(XMLConstants.XMLNS_ATTRIBUTE_NS_URI,
                "xmlns:" + bpmnMap.getNamespaceAlias(messageItemDefinition.getStructureRef().getNamespaceURI()),
                messageItemDefinition.getStructureRef().getNamespaceURI());
        propertyAliasElt.setAttribute("messageType", bpmnMap.getNamespaceAlias(messageItemDefinition.getStructureRef().getNamespaceURI()) + ":" + messageItemDefinition.getStructureRef().getLocalPart());


        List<Object> queryExpressionObjects = new ArrayList<Object>(retrievalExpression.getMessagePath().getContent());
        final ListIterator<Object> iterator = queryExpressionObjects.listIterator();
        String queryExpression = "";
        while (iterator.hasNext()) {
          final Object o = iterator.next();
          if (o instanceof String) {
            queryExpression = (String) o;
          }
        }
        Matcher matcher = Pattern.compile("^/(.*)/(.*)$").matcher(queryExpression);
        String part = null;
        String query = null;
        if (matcher.find()) {
          part = matcher.group(1);
          query = matcher.group(2);
        } else {
          Matcher matcher2 = Pattern.compile("^/(.*)$").matcher(queryExpression);
          if (matcher2.find()) {
            part = matcher2.group(1);
          } else {
            throw new TransformerException("message part not found in correlation property retrieval expression " + queryExpression);
          }
        }
        propertyAliasElt.setAttribute("part", part);
        if (query != null) {
          final Element propertyAliasQueryElt = propAliasDocument.createElementNS(WSDLGenerator.XMLNS_BPEL_2_0_VARPROP, "vprop:query");
          propertyAliasQueryElt.setTextContent(query + "/text()");
          // declare namespaces for xpath expression
          for (Entry<QName, String> otherAttribute : retrievalExpression.getMessagePath().getOtherAttributes().entrySet()) {
            if (otherAttribute.getKey().getNamespaceURI().equals(XMLConstants.XMLNS_ATTRIBUTE_NS_URI)) {
              propertyAliasQueryElt.setAttributeNS(XMLConstants.XMLNS_ATTRIBUTE_NS_URI, XMLConstants.XMLNS_ATTRIBUTE + ":" + otherAttribute.getKey().getLocalPart(), otherAttribute.getValue());
            }
          }
          propertyAliasElt.appendChild(propertyAliasQueryElt);
        }
        propAliasExtElement.setElement(propertyAliasElt);
        wsdlDefinition.addExtensibilityElement(propAliasExtElement);
      }
    }
  }
}
