/**
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.bpmn2bpel.generators;

import java.util.Collection;
import java.util.List;
import javax.xml.namespace.QName;
import org.ow2.orchestra.bpmn2bpel.BpmnMap;
import org.ow2.orchestra.bpmn2bpel.exceptions.TransformerException;
import org.ow2.orchestra.jaxb.bpel.Assign;
import org.ow2.orchestra.jaxb.bpel.Documentation;
import org.ow2.orchestra.jaxb.bpel.Reply;
import org.ow2.orchestra.jaxb.bpel.Scope;
import org.ow2.orchestra.jaxb.bpel.Sequence;
import org.ow2.orchestra.jaxb.bpel.TActivity;
import org.ow2.orchestra.jaxb.bpel.TCorrelation;
import org.ow2.orchestra.jaxb.bpel.TCorrelations;
import org.ow2.orchestra.jaxb.bpel.TInitiate;
import org.ow2.orchestra.jaxb.bpel.ToParts;
import org.ow2.orchestra.jaxb.bpel.Variable;
import org.ow2.orchestra.jaxb.bpel.Variables;
import org.ow2.orchestra.jaxb.bpmn.CorrelationKey;
import org.ow2.orchestra.jaxb.bpmn.DataInputAssociation;
import org.ow2.orchestra.jaxb.bpmn.MessageFlow;
import org.ow2.orchestra.jaxb.bpmn.TMessageEventDefinition;
import org.ow2.orchestra.jaxb.bpmn.TThrowEvent;


/**
 * @author Guillaume Porcher
 */
class ReplyGenerator {

  private final BpelGenerator bpelGenerator;

  ReplyGenerator() {
    this.bpelGenerator = new BpelGenerator();
  }

  /**
   * @param bpelGenerator
   */
  ReplyGenerator(final BpelGenerator bpelGenerator) {
    this.bpelGenerator = bpelGenerator;
  }

  TActivity createReplyActivity(final BpmnMap bpmnMap, final QName processQName,
                                final TThrowEvent event, final TMessageEventDefinition messageEventDefinition) {
    final Reply reply = new Reply();

    reply.setName(event.getId());
    final Documentation doc = new Documentation();
    doc.getContent().add(event.getName());
    reply.getDocumentations().add(doc);

    final Assign assign = new Assign();
    assign.setName(event.getId() + "_dataInput");
    final Documentation outAssignDoc = new Documentation();
    outAssignDoc.getContent().add("DataInput assignment for " + event.getName());
    assign.getDocumentations().add(outAssignDoc);

    final Scope scope = new Scope();
    scope.setName(event.getId() + "_scope");
    final Documentation scopeDoc = new Documentation();
    scopeDoc.getContent().add("Scope for " + event.getName());
    scope.getDocumentations().add(scopeDoc);

    final Variables variables = new Variables();
    scope.setVariables(variables);

    if (messageEventDefinition.getOperationRef() == null) {
      throw new TransformerException("message events must define an operation.");
    }

    final QName operationQName = messageEventDefinition.getOperationRef();


    final OperationReference replyOperation =
            this.bpelGenerator.getProcessOperationReference(bpmnMap, processQName, operationQName);
    if (replyOperation == null) {
      throw new TransformerException("operation not found");
    }
    reply.setOperation(replyOperation.getOperation().getName());
    reply.setPartnerLink(replyOperation.getPartnerLinkName());
    reply.setPortType(replyOperation.getPortTypeQName());

    // set correlations
    final Collection<MessageFlow> messageFlows = bpmnMap.getOutgoingMessageFlows(event, processQName);
    if (!messageFlows.isEmpty()) {
      if (messageFlows.size() != 1) {
        throw new TransformerException("Only one outgoing message flow supported for event " + event.getId());
      }
      final MessageFlow messageFlow = messageFlows.iterator().next();
      Collection<CorrelationKey> correlationKeys = bpmnMap.getMessageFlowCorrelations(messageFlow, processQName);
      if (!correlationKeys.isEmpty()) {
        TCorrelations correlations = new TCorrelations();
        for (CorrelationKey correlationKey : correlationKeys) {
          TCorrelation correlation = new TCorrelation();
          correlation.setInitiate(TInitiate.JOIN);
          correlation.setSet(correlationKey.getId());
          correlations.getCorrelations().add(correlation);
        }
        reply.setCorrelations(correlations);
      }
    }

    // set ToParts
    final List<DataInputAssociation> dataInputAssociations = event.getDataInputAssociations();
    if (!dataInputAssociations.isEmpty()) {
      final Object dataObject =
              this.bpelGenerator.getInputData(dataInputAssociations, replyOperation, assign, false, bpmnMap);
      if (dataObject instanceof ToParts) {
        reply.setToParts((ToParts) dataObject);
      } else if (dataObject instanceof Variable) {
        final Variable inVariable = (Variable) dataObject;
        scope.getVariables().getVariables().add(inVariable);
        reply.setVariable(inVariable.getName());
      }
    }

    if (assign.getCopiesAndExtensionAssignOperations().isEmpty()) {
      return reply;
    } else {
      final Sequence sequence = new Sequence();
      sequence.setName(event.getId() + "_sequence");
      final Documentation sequenceDoc = new Documentation();
      sequenceDoc.getContent().add("Sequence for " + event.getName());
      sequence.getDocumentations().add(sequenceDoc);
      sequence.getAssignsAndCompensatesAndCompensateScopes().add(assign);
      sequence.getAssignsAndCompensatesAndCompensateScopes().add(reply);
      if (scope.getVariables().getVariables().isEmpty()) {
        return sequence;
      } else {
        scope.setSequence(sequence);
        return scope;
      }
    }
  }

}
